#ifndef __PREDICTOR_H__
#define __PREDICTOR_H__

#include <stdlib.h>
#include "cbp3_framework.h"

#define CUMAX(x) ((1 << (x)) - 1)

//counter
//MAX : max value
//MIN : min value
template<typename T, int MAX, int MIN>
class COUNTER {
private:
	T ctr;
public:
	T read() {
		return ctr;
	}

	bool pred() {
		return ctr >= 0;
	}
	bool satmax(){
		return ctr == MAX;
	}
	bool satmin(){
		return ctr == MIN;
	}
	void write(T v) {
		assert(v <= MAX);
		assert(v >= MIN);
		ctr = v;
	}
	void add(T d) {
		ctr = ctr + d;
		if (ctr > MAX){
			ctr = MAX;
		}else if (ctr < MIN){
			ctr = MIN;
		}
	}
	void update(bool incr) {
		if (incr) {
			if (ctr < MAX)
				ctr = ctr + 1;
		} else {
			if (ctr > MIN)
				ctr = ctr - 1;
		}
	}
};
//signed integer counter
template<int WIDTH>
class SCOUNTER : public COUNTER<int32_t,((1<<(WIDTH-1))-1),(-(1<<(WIDTH-1)))>{
};
//unsigned integer counter
template<int WIDTH>
class UCOUNTER : public COUNTER<int32_t,((1<<(WIDTH))-1),0>{
};


//history register
class HISTORY {
private:
	int MAXHIST; // constant value

	bool *bhr; // branch history

public:
	void init(int HLENGTH) {
		MAXHIST = HLENGTH+1; // set constant value
		bhr = new bool[MAXHIST];
		for (int i = 0; i < MAXHIST; i++){
			bhr[i] = false;
		}
	}

	bool read(int n) {
		assert(n < MAXHIST);
		return bhr[n];
	}

	uint32_t read(int from, int n) {
		assert(n < MAXHIST);
		int r = 0;
		for (int i = from; i < n; ++i) {
			r ^= bhr[i] << ((i - from) % 32);
		}
		return r;
	}
	//push
	void push(bool t) {
		for(int i=MAXHIST-1; i>=0; i--) {
			bhr[i] = bhr[i-1];
		}
 		bhr[0] = t;
	}
};

//folded history
template<int WIDTH>
class FOLDHIST {
private:
	//path history length
	static const int PathHistoryLength = 32;

private:
	uint32_t start[2]; // constant value
	uint32_t end[2]; // constant value
	uint32_t pos[3]; // constant value

	uint32_t comp; // folded history

public:

	void init(int s, int e,int s2,int e2) {
		comp = 0;
		start[0] = s;
		start[1] = s2;
		end[0] = e;
		end[1] = e2;
		pos[0] = 0;
		pos[1] = (pos[0] + end[0] - start[0]) % WIDTH;
		pos[2] = (pos[1] + end[1] - start[1]) % WIDTH;
		assert(pos[0] < WIDTH);
		assert(pos[1] < WIDTH);
		assert(pos[2] < WIDTH);
	}
	void init(int s, int e) {
		comp = 0;
		start[0] = s;
		start[1] = s;
		end[0] = e;
		end[1] = e;
		pos[0] = 0;
		pos[1] = (pos[0] + end[0] - start[0]) % WIDTH;
		pos[2] = (pos[1] + end[1] - start[1]) % WIDTH;
		assert(pos[0] < WIDTH);
		assert(pos[1] < WIDTH);
		assert(pos[2] < WIDTH);
	}

	void init(int l) {
		init(0, l,0,(l>=PathHistoryLength)?PathHistoryLength:l);
	}

	uint32_t read(uint32_t pc) {
		assert(comp >= 0);
		assert(comp < (1 << WIDTH));

		pc &= (1 << WIDTH) - 1;
		return pc ^ comp;
	}
	uint32_t read(uint32_t pc, int rot) {
		assert(comp >= 0);
		assert(comp < (1 << WIDTH));
		uint32_t r = rot % WIDTH;
		uint32_t p = pc & ((1 << WIDTH) - 1);
		p = (p << r);
		p = p ^ (p >> WIDTH);
		p = (p & ((1 << WIDTH) - 1));
		return p ^ comp;
	}
	void update(class HISTORY &ghr, class HISTORY &phr) {
		comp = (comp << 1);
		comp |= (comp >> WIDTH);
		comp &= ((1 << WIDTH) - 1);
		comp ^= (ghr.read(start[0]) ? 1 : 0) << pos[0];
		comp ^= (ghr.read(end[0]) ? 1 : 0) << pos[1];
		comp ^= (phr.read(start[1]) ? 1 : 0) << pos[1];
		comp ^= (phr.read(end[1]) ? 1 : 0) << pos[2];
	}
};



/*********************************************
 * FILTER
 *********************************************/

/*
 * n-way SetAssociative(LRU)
 */
template<int lines,int assoc,int tagwidth>
class setassoc{
protected:
	int32_t tag[lines][assoc];
	int32_t lru[lines][assoc];
protected:
	//return line
	int getline(uint32_t key){
		return key % (lines);
	}
	//return tag
	int gettag(uint32_t key){
		uint32_t p = key / lines;
		return (p ^ (p >> tagwidth) ^ (p >> tagwidth * 2)) & ((1 <<tagwidth) - 1);
	}

	//if corresponding entry exists, return way number
	//otherwise return -1
	int hiti(uint32_t key){
		int line = getline(key);
		int t = gettag(key);
		for (int i = 0; i < assoc; ++i) {
			if(tag[line][i] == t){
				return i;
			}
		}
		return -1;
	}

	//return way number to insert
	int inserti(uint32_t key){
		int i = getline(key);
		int a = 0;
		int l = assoc;
		for(int j=0;j<assoc;++j){
			if(lru[i][j] < l){
				l = lru[i][j];
				a = j;
			}
		}
		return a;
	}
	//free entry
	void free(uint32_t key){
		int i = getline(key);
		int j = hiti(key);

		if(j < 0) return;
		int l = lru[i][j];
		for (int a = 0; a < assoc; ++a) {
			if(lru[i][a] < l){
				lru[i][a]++;
			}
		}
		lru[i][j] = 0;
	}
public:
	setassoc(){
		for (int i = 0; i < lines; ++i) {
			for (int j = 0; j < assoc; ++j) {
				tag[i][j] = 0;
				lru[i][j] = j;
			}
		}
	}


	bool hit(uint32_t key){
		return hiti(key) >=0;
	}

	//update lru
	bool updatelru(uint32_t key){
		int a = hiti(key);
		if(a == -1) return false;
		int i = getline(key);
		int l = lru[i][a];
		if(l == 0) return false;
		lru[i][a] = assoc;
		for(int j=0;j<assoc;++j){
			if(lru[i][j] > l)
				lru[i][j]--;
		}
		return true;
	}

	//update lru after insert
	bool updatelru2(uint32_t key){
		int a = hiti(key);
		if(a == -1) return false;
		int i = getline(key);
		int l = lru[i][a];
		lru[i][a] = assoc;
		for(int j=0;j<assoc;++j){
			if(lru[i][j] > l)
				lru[i][j]--;
		}
		return true;
	}

	//insert
	int insert(uint32_t key){
		int i = getline(key);
		int a = inserti(key);

		tag[i][a] = gettag(key);
		updatelru2(key);
		return a;
	}

	//insert to selected way(a)
	int insert(uint32_t key,int a){
		int i = getline(key);
		tag[i][a] = gettag(key);
		updatelru2(key);
		return a;
	}
};



/*
 * Bias filter
 *
 */
template<int tbit,bool enable>
class biaspredictor{
protected:
	static const int32_t lines = 1 << tbit;
	int32_t table[lines]; // 2bit * lines

	static const int32_t INIT = 3;
	static const int32_t TAKEN = 1;
	static const int32_t NOT_TAKEN = 0;
	static const int32_t NORMAL = 2;
public:
	biaspredictor(){
		for (int i = 0; i < lines; ++i) {
			table[i] = INIT;
		}
	}
	//index
	uint32_t idx(uint32_t pc){
		return pc ^ (pc >> tbit) ^  (pc >> (tbit*2)) ;
	}

	//if state is NORMAL then return false (not filtering)
	//otherwise return true (filtering)
	bool hit(uint32_t pc){
		pc = idx(pc);
		int t = table[pc%lines]%4;
		if(t == NORMAL){
			return false;
		}
		return enable;
	}
	//predict
	bool pred(uint32_t pc){
		pc = idx(pc);
		int t = table[pc%lines]%4;
		if(t == TAKEN){
			return true;
		}else{
			return false;
		}
	}
	//update in retire stage
	bool update(uint32_t pc,bool taken){
		pc = idx(pc);
		int t = table[pc%lines]%4;
		if(t == TAKEN){
			if(taken){
				return enable;
			}else{
				table[pc%lines] = NORMAL;
			}
		}else if(t == NOT_TAKEN){
			if(!taken){
				return enable;
			}else{
				table[pc%lines] = NORMAL;
			}
		}else if(t == INIT){
			if(taken){
				table[pc%lines] = TAKEN;
			}else{
				table[pc%lines] = NOT_TAKEN;
			}
			return enable;
		}
		return false;
	}
	int budgets(){
		return (enable)?(2)*lines :0;
	}
};


/*
 * Loop Filter
 *
 */
template<int lines, int assoc, int tagwidth, bool enable>
class looppredictor: public setassoc<lines, assoc, tagwidth> {
protected:
	static const int LC = 14;// width of loop counter
	static const int IC = LC;//width of iteration counter
	static const int SIC = IC;//width of iteration counter
	static const int CONF = 3;//initial confidence
	static const int CONFA = 7;
	static const int CONFB = 7;
	static const int LMAX = (1 << IC) - 1;
	static const int MINT = 5;

	int32_t lc[lines][assoc]; //loop counter (LC bits)
	int32_t ic[lines][assoc]; //current iteration counter that is updated in retire stage (IC bits)
	int32_t sic[lines][assoc]; //current iteration counter that is updated in fetch stage (SIC bits)
	bool d[lines][assoc]; //direction (1 bit)
	int32_t confidence[lines][assoc]; //confidence (CONF bits)
public:
	int budgets() {
		return (enable) ? lines * assoc * (LC + IC + SIC + CONF + 1 + tagwidth + 3/*LRU*/) : 0;
	}

	looppredictor() {
		for (int i = 0; i < lines; ++i) {
			for (int j = 0; j < assoc; ++j) {
				lc[i][j] = 0;
				ic[i][j] = 0;
				sic[i][j] = 0;
				d[i][j] = 0;
				confidence[i][j] = 0;
			}
		}
	}

	//predict
	bool pred(uint32_t key) {
		int i = this->hiti(key);
		if (i < 0)
			return false;
		int l = this->getline(key);
		if (lc[l][i] == LMAX) {
			return d[l][i];
		}
		if (lc[l][i] > sic[l][i]) {
			return d[l][i];
		}
		return !d[l][i];
	}

	//if the instruction seems loop return ture else return false
	bool filterhit(uint32_t key) {
		int i = this->hiti(key);
		if (i < 0)
			return false;
		int l = this->getline(key);
		int c = confidence[l][i];
		if ((((c >= CONFA && d[l][i]) || (c >= CONFB && !d[l][i])) && lc[l][i] >= MINT && lc[l][i] >= ic[l][i])
				|| lc[l][i] == LMAX) {
			return enable;
		}
		return false;
	}

	//update in retire stage
	void update(uint32_t key, bool taken) {
		int i = this->getline(key);
		int j = this->hiti(key);

		if (j < 0) {
			//allocate new entry
			int x = -1;
			//find unused entry
			for (int a = 0; a < assoc; ++a) {
				if (lc[i][a] == 0 && ic[i][a] == 0 && confidence[i][a] == 0) {
					x = a;
					break;
				}
			}
			//find low confidencd entry
			if (x < 0)
				for (int a = 0; a < assoc; ++a) {
					if (confidence[i][a] == 0) {
						x = a;
						break;
					}
				}
			//find small loop
			if (x < 0)
				for (int a = 0; a < assoc; ++a) {
					if (lc[i][a] < MINT && lc[i][a] != 0) {
						x = a;
						break;
					}
				}
			if (x >= 0) {
				x = this->insert(key, x);
			} else {
				x = this->insert(key);
			}
			lc[i][x] = 0;
			ic[i][x] = 1;
			sic[i][x] = 1;
			d[i][x] = taken;
			confidence[i][x] = 1;
		} else {
			//update
			if (d[i][j] == taken) {
				if (ic[i][j] < LMAX) {
					ic[i][j]++;
				} else {
					confidence[i][j] = CUMAX(CONF);
					lc[i][j] = LMAX;
				}
				if (ic[i][j] > lc[i][j] && lc[i][j] != 0) {
					confidence[i][j] = 0;
					this->free(key);
				}
			} else {
				if (lc[i][j] == 0) {

				} else if (lc[i][j] != ic[i][j]) {
					if (confidence[i][j] == CUMAX(CONF)) {
						confidence[i][j]--;
					} else {
						confidence[i][j] = 0;
						this->free(key);
					}
				} else {
					if (confidence[i][j] < CUMAX(CONF))
						confidence[i][j]++;
				}
				lc[i][j] = ic[i][j];
				ic[i][j] = 0;
			}
		}
		this->updatelru(key);
	}
	//update in fetch stage
	void update_fetch(uint32_t key, bool taken) {
		int i = this->getline(key);
		int j = this->hiti(key);

		if (j >= 0) {
			if (d[i][j] == taken) {
				if (sic[i][j] < LMAX) {
					sic[i][j]++;
				}
			} else {
				sic[i][j] = 0;
			}
		}
	}
};



/*
 * Exceptional Filter
 *
 * This filter detect not worth training case and stop base predictor's training such case
 */
template<int lines, int assoc, int tagwidth, bool enable>
class exceptionalfilter: public setassoc<lines, assoc, tagwidth> {
protected:
	static const uint CL = 25;//exceptional threshold

	static const uint CMAX = 31;
	static const uint CMIN = 0;

	static const uint AP = 4;//super miss
	static const uint MP = 2;//miss
	static const uint HG = 1;//hit

	// each counter contains 5 bits (0-31)
	uint32_t c[lines][assoc];
public:
	int budgets() {
		return (enable) ? lines * assoc * (5 + tagwidth + 3/*LRU*/) : 0;
	}

	exceptionalfilter() {
		for (int i = 0; i < lines; ++i) {
			for (int j = 0; j < assoc; ++j) {
				c[i][j] = 0;
			}
		}
	}

	bool filterhit(uint32_t key) {
		int i = this->hiti(key);
		if (i < 0)
			return false;
		int l = this->getline(key);
		if (c[l][i] >= CL) {
			return enable;//filter hit
		}
		return false;
	}

	//update in retire stage
	void update(uint32_t key, bool hit, bool alloc) {
		int i = this->getline(key);
		int j = this->hiti(key);
		if (j < 0) {//allocate new entry
			if(!alloc)return;
			int x = this->insert(key);
			c[i][x] = 1;
		} else {
			if (alloc) {//super miss
				if (c[i][j] <= CMAX - AP)
					c[i][j] += AP;
			} else if (!hit) {//miss
				if (c[i][j] <= CMAX - MP)
					c[i][j] += MP;
			} else {//hit
				if (c[i][j] >= CMIN + HG)
					c[i][j] -= HG;
			}
		}
		if(alloc)
			this->updatelru(key);
	}
};


/*
 * Give Up Filter
 *
 * if the branch is hard to predict (correct prediction rate is lower than its bias),
 * Give Up Filter gives prediction and stops base predictor's training
 */
template<int lines, int assoc, int tagwidth, bool enable>
class giveupfilter: public setassoc<lines, assoc, tagwidth> {
protected:
	static const uint CL = 24;
	static const uint CU = 5;
	static const uint CLT = 4;

	static const uint CMAX = 31;
	static const uint CMIN = 0;

	static const uint IC = 16;

	// each counters contains 5 bits (0-31)
	uint32_t tc[lines][assoc];//always taken gives better prediction than main predictor counter
	uint32_t ntc[lines][assoc];//always not taken gives better prediction than main predictor counter
public:
	int budgets() {
		return (enable) ? lines * assoc * (5 * 2 + tagwidth + 3/*LRU*/) : 0;
	}

	giveupfilter() {
		for (int i = 0; i < lines; ++i) {
			for (int j = 0; j < assoc; ++j) {
				tc[i][j] = 0;
				ntc[i][j] = 0;
			}
		}
	}

	//if  return true
	//else false
	bool filterhit(uint32_t key) {
		int i = this->hiti(key);
		if (i < 0)
			return false;
		int l = this->getline(key);
		if (tc[l][i] >= CL || ntc[l][i] >= CL) { //counter is eq or grater than threshold(CL)
			return enable;
		}
		return false;
	}

	//prediction
	bool pred(uint32_t key) {
		int i = this->hiti(key);
		if (i < 0)
			return false;
		int l = this->getline(key);

		if (tc[l][i] >= CL) {
			return true; //predicting always taken is better than base predictor
		}
		if (ntc[l][i] >= CL) {//predicting always not taken is better base predictor
			return false;
		}
		return false;
	}

	//filter base predictor's training or not
	bool hpred(uint32_t key) {
		int i = this->hiti(key);
		if (i < 0)
			return false;
		int l = this->getline(key);

		//Block training of main predictor if counter is eq or greater than threshold(CMAX)
		if (tc[l][i] >= CMAX) {
			return true;
		}
		if (ntc[l][i] >= CMAX) {
			return true;
		}
		return false;
	}

	//update filter
	void update(uint32_t key, bool taken, bool pred) {
		int i = this->getline(key);
		int j = this->hiti(key);
		if (j < 0) {//allocate new entry
			int x = -1;
			//search a entry that has low counter value
			if (x < 0)
				for (int a = 0; a < assoc; ++a) {
					if (tc[i][a] <= CLT && ntc[i][a] <= CLT) {
						x = a;
						break;
					}
				}
			if (x >= 0) {//found
				x = this->insert(key, x);
			} else {//not found
				x = this->insert(key);
			}
			tc[i][x] = CU;
			ntc[i][x] = CU;
		} else {
			//update counters
			if (taken) {
				if (pred == taken) {//predict is correct and outcome is taken
					if (ntc[i][j] > CMIN)
						ntc[i][j]--;
				} else {//predict is not correct and outcome is taken
					if (tc[i][j] < CMAX)
						tc[i][j]++;
				}
			} else {
				if (pred == taken) {//predict is correct and outcome is not taken
					if (tc[i][j] > CMIN)
						tc[i][j]--;
				} else {//predict is not correct and outcome is not taken
					if (ntc[i][j] < CMAX)
						ntc[i][j]++;
				}
			}
		}
		this->updatelru(key);
	}

};

#endif
